// Description: 

#ifndef __COORD_H__
#define __COORD_H__

#include <string>
//#include <iostream>
//#include <sstream>
//#include <iomanip>
#include <math.h>
#include <algorithm>
//#include <locale>

class CEllipsoid;
class CCartesianCoord;

// -------------------------------------------------------------------------
// CLASS:  CLatLon
/*! 
   \brief  Class encapsulating a latitude/longitude coordinate.

   \author fizzymagic
   \date   9/6/2003
*/
// -------------------------------------------------------------------------
class CLatLon
{
public:
   //! Default constructor.
   CLatLon() : m_Latitude(0.), m_Longitude(0.) {};
   //! Constructor.
   CLatLon(double dLat, double dLon) : m_Latitude(dLat), m_Longitude(dLon) {};

public:
   double m_Latitude;   //!< Latitude (degrees)
   double m_Longitude;  //!< Longitude (degrees)
   double m_DegDist;

   static const CEllipsoid m_Ellipsoid;   //!< Reference ellipsoid (meters).
   static const double m_Radius;          //!< Earth radius in meters.
   static const double m_Deg2Rad;         //!< Conversion factor.
   static const std::string m_strZoneLetters;   //!< UTM zones.

public:
   double SphericalDistance(CLatLon& P);
   double SphericalDistance(CLatLon& P1, CLatLon& P2);
   CLatLon SphericalProjection(double dAzimuth, double dDistance);
   double VincentyDistance(CLatLon& P);
   double VincentyDistance(CLatLon& P, double *pForwardAzimuth, double *pReverseAzimuth);
   bool IsBetween(CLatLon& P1, CLatLon& P2);
   CLatLon VincentyProjection(double dAzimuth, double dDistance);

   CCartesianCoord ToCartesian(void);
   void FromCartesian(CCartesianCoord& C);
   CCartesianCoord ToSphericalCartesian(void);
   void FromSphericalCartesian(CCartesianCoord& C);

   CCartesianCoord SphericalCross(CLatLon& P);

public:
   std::string LatToDDD(void);
   std::string LatToDMM(void);
   std::string LatToDMS(void);
   std::string LongToDDD(void);
   std::string LongToDMM(void);
   std::string LongToDMS(void);

   std::string ToDDD(void);
   std::string ToDMM(void);
   std::string ToDMS(void);
   std::string ToUTM(void);

   bool ParseCoords(const char *szCoordString);
   bool ParseCoords(std::string& strCoordString) {
      return ParseCoords(strCoordString.c_str());
   };
   bool ParseCoords(const char *szLatString, const char *szLongString);
   bool ParseCoords(std::string& strLatString, std::string& strLongString) {
      return ParseCoords(strLatString.c_str(), strLongString.c_str());
   };
   bool ParseDegreeString(std::string& strDegrees, double *dResult);
   bool ParseDegreeString(const char *szDegrees, double *dResult) {
      return ParseDegreeString(std::string(szDegrees), dResult);
   }
   bool ParseUTM(const char *szCoordString);
   bool ParseUTM(std::string& strCoordString) {
      return ParseUTM(strCoordString.c_str());
   };

   static void CleanCoordString(std::string& strCoordString);

protected:
   int GetZone(void);
   char GetZoneLetter(void);
   bool ConvertUTM(int iZone, char cZoneLetter, double dEasting, double dNorthing);
};

// -------------------------------------------------------------------------
// CLASS:  CEllipsoid
/*! 
   \brief  Class encapsulating ellipsoidal parameters for the Earth

   \author fizzymagic
   \date   9/6/2003

   All parameters should be in meters.
*/
// -------------------------------------------------------------------------
class CEllipsoid
{
public:
   //! Default constructor.  Uses WGS84 ellipsoid.
   CEllipsoid() : m_a(6378137.00), m_fInv(298.257223563) {};
   //! Constructor.
   CEllipsoid(double a, double fInv) : m_a(a), m_fInv(fInv) {};

public:
   double m_a;       //!< Semimajor axis (meters)
   double m_fInv;    //!< 1/flatness
};

// -------------------------------------------------------------------------
// CLASS:  CCartesianCoord
/*! 
   \brief  Class encapsulating xyz Cartesian coordinates.

   \author fizzymagic
   \date   9/6/2003
*/
// -------------------------------------------------------------------------
class CCartesianCoord
{
public:
   //! Default constructor.
   CCartesianCoord() : m_x(0.), m_y(0.), m_z(0.) {};
   //! Constructor from individual x, y, and z values.
   CCartesianCoord(double x, double y, double z) : m_x(x), m_y(y), m_z(z) {};
   //! Constructor from a CLatLon coordinate.  Uses the ellipsoid.
   CCartesianCoord(CLatLon& C) {
      *this = C.ToCartesian();
   };

public:
   double m_x;    //!< X coordinate.
   double m_y;    //!< Y coordinate.
   double m_z;    //!< Z coordinate.

public:
   //! Cartesian dot product.
   double dot(CCartesianCoord& C) {
      return m_x * C.m_x + m_y * C.m_y + m_z * C.m_z;
   };

   //! Cartesian cross product.
   CCartesianCoord cross(CCartesianCoord& C) {
      return CCartesianCoord(m_y * C.m_z - m_z * C.m_y, 
                             m_z * C.m_x - m_x * C.m_z, 
                             m_x * C.m_y - m_y * C.m_x);
   };

   //! Cartesian scaling.
   CCartesianCoord operator*(double R) {
      return CCartesianCoord(R * m_x, R * m_y, R * m_z);
   };
   CCartesianCoord& operator*=(double R) {
      m_x *= R; m_y *= R; m_z *= R;
      return *this;
   };
   CCartesianCoord operator/(double R) {
      return CCartesianCoord(m_x/R, m_y/R, m_z/R);
   };
   CCartesianCoord& operator/=(double R) {
      m_x /= R; m_y /= R; m_z /= R;
      return *this;
   };

   //! Addition of two Cartesian coordinates.
   CCartesianCoord operator+(CCartesianCoord& C) {
      return CCartesianCoord(m_x + C.m_x, m_y + C.m_y, m_z + C.m_z);
   };
   CCartesianCoord& operator+=(CCartesianCoord& C) {
      m_x += C.m_x; m_y += C.m_y; m_z += C.m_z;
      return *this;
   };

   //! Subtraction.
   CCartesianCoord operator-(CCartesianCoord& C) {
      return CCartesianCoord(m_x - C.m_x, m_y - C.m_y, m_z - C.m_z);
   };
   CCartesianCoord& operator-=(CCartesianCoord& C) {
      m_x -= C.m_x; m_y -= C.m_y; m_z -= C.m_z;
      return *this;
   };

   //! Negation.
   CCartesianCoord operator-(void) {
      return CCartesianCoord(-m_x, -m_y, -m_z);
   };


   //! Cartesian norm.
   double Norm(void) {
      return sqrt(m_x*m_x + m_y*m_y + m_z*m_z);
   };

   //! Cartesian normalize.
   CCartesianCoord& Normalize(void) {
      double dNorm = Norm();
      m_x /= dNorm; m_y /= dNorm; m_z /= dNorm;
      return *this;
   };
};

void findandreplace(std::string& strSource, std::string& strFind, std::string& strReplace);
void trim(std::string& strSource);

#endif   // __COORD_H__
